/*
    mpg321 - a fully free clone of mpg123.
    Copyright (C) 2001 Joe Drew
    
    Originally based heavily upon:
    plaympeg - Sample MPEG player using the SMPEG library
    Copyright (C) 1999 Loki Entertainment Software
    
    Also uses some code from
    mad - MPEG audio decoder
    Copyright (C) 2000-2001 Robert Leslie
    
    Original playlist code contributed by Tobias Bengtsson <tobbe@tobbe.nu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


   --
    Changes:
    - Make sure to print @P 0 from read_from_fd at the end
    - "Fix" read_from_fd() audio glitching
    - added "SOFTWARE_VOLUME" flag to enable/disable software volume adjustment
      (vs. using the driver)
    - handle errors more gracefully than exit()ing, handle failures in
      open_ao_playdevice by BREAKing the decoder
    - Use libmad's "resample" code to do sample rate conversion if opening the 
      ao_playdevice fails.
    - Change read_from_fd to handle non-blocking reads, and return errors to
      indicate status of the network, including a timeout if network doesn't
      come back.  Also allow cancelling this "wait" on stdin activity.
    - Added seek_by_time functionality
    - Added limit to how much time we'll spend scanning to calculate frame
      count and duration
    by TiVo Inc. (2002,2003)
   --
*/

#include <unistd.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>

#include <sys/socket.h>

#include "mpg321.h"

#ifdef USE_MAD_RESAMPLE
#include "resample.h"
/*#define FORCE_RESAMPLE*/
#else
#undef FORCE_RESAMPLE
#endif

// When scanning a file to calculate the duration, we bail if it takes longer
// than this many seconds.  This can happen on VERY large files, say, on a CD.
#define SCAN_TIMEOUT_SECS    2

// The maximum size (in bytes) of a file we'll even *try* to calculate the
// duration of.
#define MAX_FILE_SIZE_FOR_DURATION_CALC       Mb(10)-1

// The number of frames we check from the beginning of the file for variable
// bitrate (VBR).  The higher this number is, the more accurate our guess will
// be regarding whether a file is VBR or not (but also the longer it will take
// to make the guess)
#define NUM_FRAMES_FOR_VBR_ESTIMATE     20

//#define TRACE_READ_FROM_FD    1
//#define TRACE_READ_FROM_FD_2    1

unsigned long current_frame=0;

enum mad_flow read_from_mmap(void *data, struct mad_stream *stream)
{
    buffer *playbuf = (buffer *)data;
    void *mpegdata = NULL;
    
    /* libmad asks us for more data when it runs out. We don't have any more,
       so we want to quit here. */
    if (status != MPG321_REWINDING && 
        status != MPG321_SEEKING &&
        playbuf->done)
    {
        mpg321_printf("%s\n", "Done");
        status = MPG321_STOPPED;
        mpg321_msg("@P 0\n");
        return MAD_FLOW_STOP;
    }

    mpg321_printf("%s\n", "Start");
    
    if(playbuf->fd != -1)
    {
        mpg321_error("read_from_mmap called when not expected!");
        return MAD_FLOW_BREAK;
    }
    
    mpegdata = playbuf->buf;

    if(status == MPG321_REWINDING)
    {
        mpegdata = playbuf->frames[current_frame];
        options.seek = 0;
        status = MPG321_PLAYING;
    }
    
    if (status == MPG321_SEEKING && playbuf->seek_by_time_byte_offset)
    {
        mpg321_printf("Seeking to %ld\n", playbuf->seek_by_time_byte_offset);
        mpegdata = playbuf->buf + playbuf->seek_by_time_byte_offset;
        playbuf->seek_by_time_byte_offset = 0;
        status = MPG321_PLAYING;
    }

    if (status != MPG321_SEEKING) /* seeking goes to playing during the decoding process */
        status = MPG321_PLAYING;
    
    playbuf->done = 1;

    mad_stream_buffer(stream, mpegdata, playbuf->length - (mpegdata - playbuf->buf));
    
    return MAD_FLOW_CONTINUE;
}

/* assumes that playbuf->buf has been preallocated to BUF_SIZE */
enum mad_flow read_from_fd(void *data, struct mad_stream *stream)
{
    buffer *playbuf = data;
    int len = 0;
    size_t bytes_to_move = 0;
    extern ssize_t fd_read_timeout(int, void*, size_t);
    void  *buf_ptr = NULL;
    size_t num_bytes_to_read = 0;
    size_t num_bytes_read = 0;
    static int bufDecodeSize = 0;
     
    if(playbuf->done)
    {
        mpg321_msg("@P 0\n");
        status = MPG321_STOPPED;
        return MAD_FLOW_STOP;
    }

    if(playbuf->fd == -1)
    {
        //XXX???Sometimes we get a spurious error here when we interrupt
        //XXX   playback to start a new song.  Suppressing until we can 
        //XXX   figure out why 
        // mpg321_error("%s called when not expected!", __FUNCTION__);
        return MAD_FLOW_BREAK;
    }
    
    if ( 0 == bufDecodeSize )
    {
        int x = sizeof(bufDecodeSize);
        // NOTE: This getsockopt() will definitely fail if playbuf->fd is not
        // socket...that's on purpose.
        if ( getsockopt( playbuf->fd,
                         SOL_SOCKET, SO_RCVBUF, &bufDecodeSize, &x ) < 0 ) 
        {
            bufDecodeSize = BUF_DECODE_SIZE;
        }
        else {
            // Half as large means we'll
            // fill the input buffer 
            bufDecodeSize = bufDecodeSize / 2; 
        }
    }
                         
    if(!playbuf->buf)
    {
        mpg321_error("%s called with null buf!", __FUNCTION__);
    }
    
    if (status == MPG321_SEEKING && playbuf->seek_by_time_byte_offset)
    {
        mpg321_printf("Seeking to %ld\n", playbuf->seek_by_time_byte_offset);
        lseek(playbuf->fd, playbuf->seek_by_time_byte_offset, SEEK_SET );
        status = MPG321_PLAYING;
    }
    
    if (stream->next_frame) {
        bytes_to_move = (long)playbuf->buf + playbuf->input_length -
            (long)stream->next_frame ;
        if ( (long)bytes_to_move > 0 ) 
        {
            // XXX It is unclear if this is an error 
//              if ( ((long)stream->next_frame - (long)playbuf->buf) < (long)bytes_to_move )
//              {
//                  mpg321_error("Moving bytes overlaps by %ld!",
//                               bytes_to_move - ((long)stream->next_frame -
//                                                (long)playbuf->buf));
//              }
            
#ifdef TRACE_READ_FROM_FD_2
            mpg321_printf("MEMMOVING: buf: %p, input_length: %ld next_frame: %p btm: %d\n", 
                          playbuf->buf, playbuf->input_length, stream->next_frame, 
                          bytes_to_move);
#endif
            
            if ( bytes_to_move > BUF_SIZE )
            {
                mpg321_error("Internal error: bytes_to_move = %d (%#x) > BUF_SIZE(%d)\n",
                             bytes_to_move, bytes_to_move, BUF_SIZE);
                return MAD_FLOW_BREAK;
            }
            memmove(playbuf->buf, stream->next_frame, bytes_to_move);
            playbuf->input_length = bytes_to_move;
        }
    }

#define READ_AS_MUCH_AS_POSSIBLE    
    buf_ptr = playbuf->buf + playbuf->input_length;
    num_bytes_to_read = BUF_SIZE - playbuf->input_length;
    do {
#ifdef TRACE_READ_FROM_FD
        mpg321_printf(">>> Reading%s %d bytes into buffer %p at %p size %ld\n",
                      ( playbuf->input_length < INPUT_BUFFER_LOW_WATER_MARK ) ?
                      "[MAY BLOCK]":"",
                      num_bytes_to_read, playbuf->buf, buf_ptr,
                      playbuf->input_length );
#endif        
        if ( playbuf->input_length < INPUT_BUFFER_LOW_WATER_MARK )
        {
            /* This is a non-blocking read, unless there's a network delay, in
               which case it will block */
            len = fd_read_timeout(playbuf->fd, buf_ptr, num_bytes_to_read );
        }
        else {
            /* This will be a non-blocking read: */
            len = read(playbuf->fd, buf_ptr, num_bytes_to_read);
        }

        if ( len > 0 ) {
            buf_ptr += len;
            num_bytes_to_read -= len;
            num_bytes_read += len;
        }
            
    }
    while ( ( len == -1 && errno == EINTR )
#ifdef READ_AS_MUCH_AS_POSSIBLE            
            || ( len > 0 && num_bytes_to_read > 0 )
#endif            
        );

#ifdef TRACE_READ_FROM_FD_2
    mpg321_printf("<<< Read %d bytes (num_bytes_read %d, num_bytes_to_read %d)\n",
                  len, num_bytes_read, num_bytes_to_read );
#endif
    if ( len < 0 ) {
        assert( len == -1 );

        if ( errno == EAGAIN ) {
#ifdef TRACE_READ_FROM_FD
            mpg321_printf("%s: CONTINUING... (%s)\n", __FUNCTION__, strerror(errno) );
#endif
            if ( 0 == num_bytes_read ) {
                mpg321_error("%s: read() failed (network lost?)", __FUNCTION__ );
            }
            
            len = 0;
        } 
        else {
            mpg321_error("%s: read() returned -1", __FUNCTION__ );
            playbuf->done = 1;
            return MAD_FLOW_BREAK;
        }
    }
    else if (len == 0) {
        
#ifdef TRACE_READ_FROM_FD
        mpg321_printf("=== DONE\n");
#endif        
        
        playbuf->done = 1;

        assert(BUF_SIZE - playbuf->input_length >= MAD_BUFFER_GUARD);

        while (len < MAD_BUFFER_GUARD)
            *((char*)playbuf->buf + playbuf->input_length + len++) = 0;
    }
    
    playbuf->input_length += num_bytes_read;

    {
        /*
         * If we are near the end of the buffer (playbuf->done), send all of it
         * to the decoder.
         * Otherwise, send only up to a "maximum" ceiling equal to some
         * fraction of the input I/O buffer.
         */
        int howMuchToSend =  (playbuf->done)?playbuf->input_length:
            MIN( bufDecodeSize, playbuf->input_length );

        mad_stream_buffer(stream, playbuf->buf, howMuchToSend);
    
#ifdef TRACE_READ_FROM_FD
        fprintf( stderr,
                 "BUFFER: %ld, (read %d bytes, copied %d, %d bytes sent to decoder)\n"
                 "----------------------------------------------------------------------------\n",
                 playbuf->input_length, len, bytes_to_move, howMuchToSend );
#endif
    }
    
    mpg321_msg("@E mpg321 buffer status:%ld\n",
               ((long)100 * (long)playbuf->input_length)/((long)BUF_SIZE));
    
    return MAD_FLOW_CONTINUE;
}    

char * layerstring(enum mad_layer layer)
{
    switch(layer)
    {
        case MAD_LAYER_I:
            return "I";
        case MAD_LAYER_II:
            return "II";
        case MAD_LAYER_III:
            return "III";
        default:
            return "?";
    }
}
    
char * modestring(enum mad_mode mode)
{
    switch(mode)
    {
        case MAD_MODE_SINGLE_CHANNEL:
            return "mono";
        case MAD_MODE_DUAL_CHANNEL:
            return "dual-channel";
        case MAD_MODE_JOINT_STEREO:
            return "joint-stereo";
        case MAD_MODE_STEREO:
            return "stereo";
        default:
            return "?";
    }
}

char * modestringucase(enum mad_mode mode)
{
    switch(mode)
    {
        case MAD_MODE_SINGLE_CHANNEL:
            return "Single-Channel";
        case MAD_MODE_DUAL_CHANNEL:
            return "Dual-Channel";
        case MAD_MODE_JOINT_STEREO:
            return "Joint-Stereo";
        case MAD_MODE_STEREO:
            return "Stereo";
        default:
            return "?";
    }
}

enum mad_flow read_header(void *data, struct mad_header const * header)
{
    char long_currenttime_str[14]; /* this *will* fill if you're using 100000+ minute mp3s */
    char long_remaintime_str[14];
    
    buffer *playbuf = (buffer *)data;
    mad_timer_t time_remaining;

    if (stop_playing_file)
    {
        errno = EIO;
        mpg321_error("Error reading file");
        stop_playing_file = 0;
        return MAD_FLOW_STOP;
    }
    
#ifdef OLD_PAUSE_IMPL
    if(options.opt & MPG321_REMOTE_PLAY)
    {
        enum mad_flow mf;

        /* We might have to stop if the user inputs something */
        if ((mf = remote_get_input_nowait(playbuf)))
            return mf;
    }
#endif

    /* Stop playing if -n is used, and we're at the frame specified. */
    if (((long)playbuf->max_frames != -1) && (current_frame > playbuf->max_frames))
    {
      status = MPG321_STOPPED;
      return MAD_FLOW_STOP;
    }

    current_frame++;

    mad_timer_add(&current_time, header->duration);

    if(options.opt & (MPG321_VERBOSE_PLAY | MPG321_REMOTE_PLAY))
    {
        mad_timer_string(current_time, long_currenttime_str, "%.2u:%.2u.%.2u", MAD_UNITS_MINUTES,
                            MAD_UNITS_CENTISECONDS, 0);

        if (mad_timer_compare(playbuf->duration, mad_timer_zero) == 0)
            time_remaining = current_time;
        else
            time_remaining = playbuf->duration;

        mad_timer_negate(&current_time);

        mad_timer_add(&time_remaining, current_time);
        mad_timer_negate(&current_time);

        mad_timer_string(time_remaining, long_remaintime_str, "%.2u:%.2u.%.2u", MAD_UNITS_MINUTES,
                            MAD_UNITS_CENTISECONDS, 0);
    }
                        
    /* update cached table of frames & times */
    if (current_frame <= playbuf->num_frames) /* we only allocate enough for our estimate. */
    {
        playbuf->frames[current_frame] = playbuf->frames[current_frame-1] + (header->bitrate / 8 / 1000)
            * mad_timer_count(header->duration, MAD_UNITS_MILLISECONDS);
        playbuf->times[current_frame] = current_time;
    }
    
    if (file_change)
    {
        file_change = 0;
        mpg321_msg("@S 1.0 %d %d %s %d %ld %d %d %d %d %ld %d %ld\n",
                   header->layer, header->samplerate, 
                   modestringucase(header->mode), header->mode_extension, 
                   (header->bitrate / 8 / 100) *
    mad_timer_count(header->duration, MAD_UNITS_CENTISECONDS),
                   /* TIVO */
                   MAD_NCHAN_WOPT(options, header),
                   header->flags & MAD_FLAG_COPYRIGHT ? 1 : 0, 
                   header->flags & MAD_FLAG_PROTECTION ? 1 : 0, header->emphasis,
                   header->bitrate/1000, header->mode_extension,
                   mad_timer_count(playbuf->duration, MAD_UNITS_MILLISECONDS)
                   );
        
        if (options.opt & MPG321_VERBOSE_PLAY)/*zip it good*/
        {
            fprintf(stderr, "MPEG 1.0, Layer: %s, Freq: %d, mode: %s, modext: %d, BPF : %ld\n"
                    "Channels: %d, copyright: %s, original: %s, CRC: %s, emphasis: %d.\n"
                    "Bitrate: %ld Kbits/s, Extension value: %d\n"
                    "Audio: 1:1 conversion, rate: %d, encoding: signed 16 bit, channels: %d\n"
                    "Duration: %ld ms\n",
                    layerstring(header->layer), header->samplerate, modestringucase(header->mode), header->mode_extension, 
                    (header->bitrate / 100) * mad_timer_count(header->duration, MAD_UNITS_CENTISECONDS),
                    /* TIVO */
                    MAD_NCHAN_WOPT(options, header),
                    header->flags & MAD_FLAG_COPYRIGHT ? "Yes" : "No",
                    header->flags & MAD_FLAG_ORIGINAL ? "Yes" : "No", header->flags & MAD_FLAG_PROTECTION ? "Yes" : "No",
                    header->emphasis, header->bitrate/1000, header->mode_extension,
                    header->samplerate,
                    /* TIVO */
                    MAD_NCHAN_WOPT(options, header),
                    mad_timer_count(playbuf->duration, MAD_UNITS_MILLISECONDS)
                );
        }

        else if (!(options.opt & MPG321_QUIET_PLAY))/*I love Joey*/
        {
            fprintf(stderr, "MPEG 1.0 layer %s, %ld kbit/s, %d Hz %s\n",
                layerstring(header->layer), header->bitrate/1000, header->samplerate, modestring(header->mode));
        }
    }
    
    if (status == MPG321_SEEKING && options.seek)
    {
        if (!--options.seek)
            status = MPG321_PLAYING;

        return MAD_FLOW_IGNORE;
    }
#ifdef OLD_PAUSE_IMPL
    else
#else  /*  OLD_PAUSE_IMPL */
    else if ( status != MPG321_PAUSED )
#endif  /*  OLD_PAUSE_IMPL */
    {
        status = MPG321_PLAYING;
    }

    if (options.opt & MPG321_VERBOSE_PLAY)
    {
        if (!options.skip_printing_frames 
            || (options.skip_printing_frames && !(current_frame % options.skip_printing_frames)))
            fprintf(stderr, "Frame# %5lu [%5lu], Time: %s [%s], \r", current_frame, 
                    playbuf->num_frames > 0 ? playbuf->num_frames - current_frame : 0, long_currenttime_str, long_remaintime_str);
    }
    
    else if (options.opt & MPG321_REMOTE_PLAY)
    {
        if (!options.skip_printing_frames 
            || (options.skip_printing_frames && !(current_frame % options.skip_printing_frames)))
            mpg321_msg("@F %ld %ld %.2f %.2f\n", current_frame, playbuf->num_frames - current_frame,
                ((double)mad_timer_count(current_time, MAD_UNITS_CENTISECONDS)/100.0),
                ((double)mad_timer_count(time_remaining, MAD_UNITS_CENTISECONDS)/100.0));
    }
    
    return MAD_FLOW_CONTINUE;
}        

/* XING parsing is from the MAD winamp input plugin */

struct xing {
  int flags;
  unsigned long frames;
  unsigned long bytes;
  unsigned char toc[100];
  long scale;
};

enum {
  XING_FRAMES = 0x0001,
  XING_BYTES  = 0x0002,
  XING_TOC    = 0x0004,
  XING_SCALE  = 0x0008
};

# define XING_MAGIC     (('X' << 24) | ('i' << 16) | ('n' << 8) | 'g')

static
int parse_xing(struct xing *xing, struct mad_bitptr ptr, unsigned int bitlen)
{
  if (bitlen < 64 || mad_bit_read(&ptr, 32) != XING_MAGIC)
    goto fail;

  xing->flags = mad_bit_read(&ptr, 32);
  bitlen -= 64;

  if (xing->flags & XING_FRAMES) {
    if (bitlen < 32)
      goto fail;

    xing->frames = mad_bit_read(&ptr, 32);
    bitlen -= 32;
  }

  if (xing->flags & XING_BYTES) {
    if (bitlen < 32)
      goto fail;

    xing->bytes = mad_bit_read(&ptr, 32);
    bitlen -= 32;
  }

  if (xing->flags & XING_TOC) {
    int i;

    if (bitlen < 800)
      goto fail;

    for (i = 0; i < 100; ++i)
      xing->toc[i] = mad_bit_read(&ptr, 8);

    bitlen -= 800;
  }

  if (xing->flags & XING_SCALE) {
    if (bitlen < 32)
      goto fail;

    xing->scale = mad_bit_read(&ptr, 32);
    bitlen -= 32;
  }

  return 1;

 fail:
  xing->flags = 0;
  return 0;
}


/* Following two functions are adapted from mad_timer, from the 
   libmad distribution */
void scan(void const *ptr, ssize_t len, buffer *buf)
{
    struct mad_stream stream;
    struct mad_header header;
    struct xing xing;
    struct timeval timestart;
    struct timezone tz;
    
    unsigned long bitrate = 0;
    int has_xing = 0;
    int is_vbr = 0;

    mad_stream_init(&stream);
    mad_header_init(&header);

    mad_stream_buffer(&stream, ptr, len);

    buf->num_frames = 0;

    /* There are three ways of calculating the length of an mp3:
      1) Constant bitrate: One frame can provide the information
         needed: # of frames and duration. Just see how long it
         is and do the division.
      2) Variable bitrate: Xing tag. It provides the number of 
         frames. Each frame has the same number of samples, so
         just use that.
      3) All: Count up the frames and duration of each frames
         by decoding each one. We do this if we've no other
         choice, i.e. if it's a VBR file with no Xing tag.
         **TIVO**
         don't bother if file is larger than a certain size
         **TIVO**
     */

/** TIVO **/
#if ( defined(SCAN_TIMEOUT_SECS) && ( SCAN_TIMEOUT_SECS > 0 ) )
    timestart.tv_sec = timestart.tv_usec = 0L;
    gettimeofday(&timestart, &tz);
#endif    
/** TIVO **/

    while (1)
    {
        if (mad_header_decode(&header, &stream) == -1)
        {
            if (MAD_RECOVERABLE(stream.error))
                continue;
            else
                break;
        }

        /* Limit xing testing to the first frame header */
        if (!buf->num_frames++)
        {
            if(parse_xing(&xing, stream.anc_ptr, stream.anc_bitlen))
            {
                is_vbr = 1;
                
                if (xing.flags & XING_FRAMES)
                {
                    /* We use the Xing tag only for frames. If it doesn't have that
                       information, it's useless to us and we have to treat it as a
                       normal VBR file */
                    has_xing = 1;
                    buf->num_frames = xing.frames;
                    break;
                }
            }
        }                

        /* Test the first n frames to see if this is a VBR file */
        if (!is_vbr && !(buf->num_frames > NUM_FRAMES_FOR_VBR_ESTIMATE))
        {
            if (bitrate && header.bitrate != bitrate) {
                is_vbr = 1;
            } else {
                bitrate = header.bitrate;
            }
        }
        
        /* We have to assume it's not a VBR file if it hasn't already been
           marked as one and we've checked n frames for different bitrates */
        else if (!is_vbr)
        {
            mpg321_printf("%s\n", "Assuming CBR");
            break;
        }
        
        mad_timer_add(&buf->duration, header.duration);

/** TIVO **/
#ifdef MAX_FILE_SIZE_FOR_DURATION_CALC
        if ( is_vbr &&
             buf->num_frames > NUM_FRAMES_FOR_VBR_ESTIMATE && 
             len > MAX_FILE_SIZE_FOR_DURATION_CALC )
        {
            if (options.opt & MPG321_REMOTE_PLAY) {
                mpg321_msg("@E MAD error:0: VBR file too big to scan for duration.\n");
            }
            else {
                mpg321_printf("%s", "mpg321: VBR file too big to scan for duration.\n");
            }
            buf->num_frames = 0;
            mad_timer_reset(&buf->duration);
            break;
        }
#endif
#if ( defined(SCAN_TIMEOUT_SECS) && ( SCAN_TIMEOUT_SECS > 0 ) )
        /* Every 1000 frames, check the timeout */
        if ( 0 == ( buf->num_frames % 1000 ) )
        {
            struct timeval timeout;
            int s = 0;
            timeout.tv_sec = timeout.tv_usec = 0L;
            s = gettimeofday( &timeout, &tz );
//             fprintf(stderr, "mpg321: checking time: %ld secs until timeout (frame %ld)\n",
//                     SCAN_TIMEOUT_SECS - (timeout.tv_sec - timestart.tv_sec), buf->num_frames);
            if ( 0 == s &&
                 ( timeout.tv_sec - timestart.tv_sec ) >= SCAN_TIMEOUT_SECS )
            {
                /* Bail on calculating duration & number of frames.  Since this
                 * is probably a VBR file, we can't even really let the code
                 * below estimate the duration & num_frames.  So, we just give
                 * up and zero out everything. 
                 */
                if (options.opt & MPG321_REMOTE_PLAY) {
                    mpg321_msg("@E MAD error:0: Bailing on duration calculation ... took too long!\n");
                }
                else {
                    fprintf( stderr, "mpg321: Bailing on duration calculation ... took too long!\n");
                }
                buf->num_frames = 0;
                mad_timer_reset(&buf->duration);
                break;
            }
        }
#endif        
/** TIVO **/
    }

    if (!is_vbr)
    {
        double time = (len * 8.0) / (header.bitrate); /* time in seconds */
        double timefrac = (double)time - ((long)(time));
        long nsamples = 32 * MAD_NSBSAMPLES(&header); /* samples per frame */
        
        /* samplerate is a constant */
        buf->num_frames = (long) (time * header.samplerate / nsamples);

        mad_timer_set(&buf->duration, (long)time, (long)(timefrac*100), 100);
    }
        
    else if (has_xing)
    {
        /* modify header.duration since we don't need it anymore */
        mad_timer_multiply(&header.duration, buf->num_frames);
        buf->duration = header.duration;
    }

    else
    {
        /* the durations have been added up, and the number of frames
           counted. We do nothing here. */
    }
    
    mad_header_finish(&header);
    mad_stream_finish(&stream);
}

#ifdef OLD_PAUSE_IMPL    
void pause_play(buffer *buf, playlist *pl)
{
    static char file[PATH_MAX] = "";
    static signed long seek = 0;
    
    if (buf == NULL && pl == NULL) /* reset */
    {
        strcpy(file, "");
        seek = 0;
        return;
    }
    
    /* pause */
    if (strlen(file) == 0)
    {
        status = MPG321_PAUSED;
        strncpy(file, buf->filename, PATH_MAX);
        file[PATH_MAX-1]='\0';
        clear_remote_file(pl);
        seek = current_frame;
        mpg321_msg("@P 1\n");
    }
    /* unpause */
    else
    {
        status = MPG321_SEEKING;
        play_remote_file(pl, file);
        file[0] = '\0';
        options.seek = seek;
        current_frame = 0;
        mpg321_msg("@P 2\n");
    }
}
#else /* OLD_PAUSE_IMPL */
void pause_play()
{
    /* New pause implementation that doesn't stop input */
    
    /* pause */
    if ( status != MPG321_PAUSED )
    {
        status = MPG321_PAUSED;
        mpg321_printf("%s\n", "PAUSED");
        if ( playdevice ) {
            if ( ! ao_stop(playdevice) ) {
                mpg321_error( "ioctl error" );
            }
        }
        mpg321_msg("@P 1\n");
    }
    /* unpause */
    else
    {
        status = MPG321_SEEKING;
        mpg321_printf("%s\n", "UNPAUSED");
        mpg321_msg("@P 2\n");
    }
}
void stop_play(buffer *buf)
{
    buf->input_length = 0;
}
#endif /* OLD_PAUSE_IMPL */    

/* seek to absolute frame frame */
void seek(buffer *buf, signed long frame)
{
    if (frame > (long)buf->num_frames)
        options.seek = buf->num_frames;
    else
        options.seek = frame;
    current_frame = 0;
    status = MPG321_SEEKING;
}

enum mad_flow move(buffer *buf, signed long frames)
{
    if (frames == 0)
        return 0;
    
    /* Our normal skipping (for -k) code handles forward seeks.
       Rewinds are handled by a stop in decoding, a rewind, and
       a restart in decoding, implemented in the main loop and in the other
       functions */
    if ((frames + current_frame) > buf->num_frames)
        options.seek = buf->num_frames - current_frame;
    else
        options.seek = frames;

    status = MPG321_SEEKING;

    /* Rewinding doesn't correct for frames on its own, so we must do so */
    if (frames < 0)
    {
        status = MPG321_REWINDING;
        if (((signed long)current_frame + frames) < 0)
            current_frame = 0;
        else
            current_frame += frames;
        current_time = buf->times[current_frame];
        return MAD_FLOW_STOP;
    }

    return 0;
}

/** TIVO **/
/* seek to absolute time */
/* If the user performs a trickplay, mpg321 does a 
 * seek_by_time by looking at the first 20 frames.  If in
 * those 20 frames we don't see a different bitrate, then 
 * we assume the file is CBR, and we make an estimate as 
 * to the byte-offset to jump to in the file.  We do this 
 * by re-initializing the decoder and starting at that 
 * offset (NOTE this only will work in MMAP mode!!).
 * If, however, seek_by_time determines the file is VBR,
 * then we seek frame-by-frame for the desired time.
 * (see scan() for similar functionality) 
 */
void seek_by_time(buffer *buf, const mad_timer_t time_to_seek )
{
    struct mad_stream stream;
    struct mad_header header;
    long frame = 0;
    long bytes = 0;
    mad_timer_t cur_time;
#ifdef DEBUG_TIME_SEEK_BY_TIME
    struct timeval timestart, timeend;
    struct timezone tz;
#endif /* DEBUG_TIME_SEEK_BY_TIME */
    
    unsigned long bitrate = 0;
    int is_vbr = 0;
    int sent_error = 0;
    
#ifdef DEBUG_TIME_SEEK_BY_TIME
    timestart.tv_sec = timestart.tv_usec = 0L;
    gettimeofday(&timestart, &tz);
#endif /* DEBUG_TIME_SEEK_BY_TIME */

    mpg321_printf( "seeking to %lu seconds into song (total = %lu frames)\n",
                   time_to_seek.seconds, buf->num_frames);
    
    mad_timer_reset(&cur_time);

    mad_stream_init(&stream);
    mad_header_init(&header);

    mad_stream_buffer(&stream, buf->buf, buf->length);

    while ( mad_timer_compare( cur_time, time_to_seek) <= 0 )
    {
        if (mad_header_decode(&header, &stream) == -1)
        {
            if (MAD_RECOVERABLE(stream.error)) {
                continue; 
            }
            else {
                frame = -1;
                break;
            }
        }

        /* Test the first n frames to see if this is a CBR file.  If it is,
         * stop scanning and calculate the frame we should go to */
        if (!is_vbr && !(frame > NUM_FRAMES_FOR_VBR_ESTIMATE))
        {
            if (bitrate && header.bitrate != bitrate) {
                mpg321_printf("VBR file! (frame %ld)\n", frame);
                is_vbr = 1;
            } else {
                bitrate = header.bitrate;
            }
        }
        else if (!is_vbr)
        {
#ifdef DEBUG_TIME_SEEK_BY_TIME
//#define FORCE_ACCURATE_SEEK_BY_TIME
#endif            
#ifndef FORCE_ACCURATE_SEEK_BY_TIME
            /* We're beyond N frames of a (potential) CBR file...break */
            mpg321_printf("%s", "Assuming CBR\n");
            break;
#endif /* FORCE_ACCURATE_SEEK_BY_TIME */
        }

        if (is_vbr && ! sent_error) 
        {
            // Fake a read failure, until we're done seeking
            errno = EAGAIN;
            mpg321_error("%s: seeking...", __FUNCTION__ );
            sent_error = 1;
        }
        
        mad_timer_add(&cur_time, header.duration);
        frame++;
    }

    bytes = stream.next_frame - stream.buffer;
    
    mad_header_finish(&header);
    mad_stream_finish(&stream);

#ifdef FORCE_ACCURATE_SEEK_BY_TIME
    mpg321_printf( "accurate frame %ld (bytes: %ld)\n", frame, bytes);
#endif /* FORCE_ACCURATE_SEEK_BY_TIME */
    
    if ( ! is_vbr 
#ifndef FORCE_ACCURATE_SEEK_BY_TIME
         && mad_timer_compare(cur_time, time_to_seek) <= 0 
#endif /* FORCE_ACCURATE_SEEK_BY_TIME */
        )
    {
        /* We stopped scanning a CBR file to make an estimate...*/
        long est_frame = 0;
        long samples_per_frame = 32 * MAD_NSBSAMPLES(&header);
        
        /* samplerate is a constant */
        long samples = mad_timer_count( time_to_seek, (enum mad_units)header.samplerate);

        est_frame = samples / samples_per_frame + 1;
        mpg321_printf( "(CBR estimate) est_frame= %lu, samplerate=%d, samples: %ld\n", 
                   est_frame, header.samplerate, samples);
        
#ifndef FORCE_ACCURATE_SEEK_BY_TIME
        frame = est_frame;
        // XXX Is MAD_UNITS_SECONDS accurate enough?
        bytes = ( header.bitrate * mad_timer_count( time_to_seek, MAD_UNITS_SECONDS ) ) / 8;
#endif /* FORCE_ACCURATE_SEEK_BY_TIME */
    }
    
    mpg321_printf( "seeking to frame %ld (byte %ld)\n",frame, bytes);
    
    if ( frame >= 0 ) {
#ifdef SEEK_FRAME_BY_FRAME
        seek(buf, frame);
#else  /* SEEK_FRAME_BY_FRAME */
        buf->seek_by_time_byte_offset = bytes;
        mpg321_printf("seek_by_time_byte_offset = %ld\n", buf->seek_by_time_byte_offset);
        
        current_frame = frame;
        mad_timer_reset(&current_time);
        mad_timer_add(&current_time, time_to_seek);//absolute time!
        
        status = MPG321_SEEKING;
#endif /* SEEK_FRAME_BY_FRAME */
    }
    else {
//        mpg321_error("couldn't seek_by_time");
        mpg321_msg("@E MAD error:%#x: couldn't seek_by_time: %s\n",
                   stream.error, mad_stream_errorstr(&stream)); 
    }

#ifdef DEBUG_TIME_SEEK_BY_TIME
    gettimeofday(&timeend, &tz);
    mpg321_printf( "took %lu.%06lu secs\n", 
               timeend.tv_sec - timestart.tv_sec,
               timeend.tv_usec - timestart.tv_usec);
#endif /* DEBUG_TIME_SEEK_BY_TIME */
}
/** TIVO **/

int calc_length(char *file, buffer *buf)
{
    int f;
    struct stat filestat;
    void *fdm;
    char buffer[3];
    int map_length = 0;
    size_t pagesize = getpagesize();

    f = open(file, O_RDONLY);

    if (f < 0)
    {
        mpg321_error("%s: couldn't open <<<%s>>>", __FUNCTION__, file);
        return -1;
    }

    if (fstat(f, &filestat) < 0)
    {
        mpg321_error("%s: couldn't stat <<<%s>>>", __FUNCTION__, file);
        close(f);
        return -1;
    }

    if (!S_ISREG(filestat.st_mode))
    {
        mpg321_error("%s: <<<%s>>> is not a regular file\n", __FUNCTION__, file);
        close(f);
        return -1;
    }

    /* TAG checking is adapted from XMMS */
    buf->length = filestat.st_size;

    if (lseek(f, -128, SEEK_END) < 0)
    {
        /* File must be very short or empty. Forget it. */
        close(f);
        return -1;
    }    

    if (read(f, buffer, 3) != 3)
    {
        close(f);
        return -1;
    }
    
    if (!strncmp(buffer, "TAG", 3))
    {
        buf->length -= 128; /* Correct for id3 tags */
    }

    map_length = (( buf->length + (pagesize - 1) ) / pagesize ) * pagesize;

    fdm = mmap(0, map_length, PROT_READ, MAP_SHARED, f, 0);
    if (fdm == MAP_FAILED)
    {
        mpg321_error("mmap failed on <<<%s>>>", file);
        close(f);
        return -1;
    }

    /* Scan the file for a XING header, or calculate the length,
       or just scan the whole file and add everything up. */
    scan(fdm, buf->length, buf);

    if (munmap(fdm, map_length) == -1)
    {
        mpg321_error("munmap failed on <<<%s>>>", file);
        close(f);
        return -1;
    }

    if (close(f) < 0)
    {
        mpg321_error("close failed on <<<%s>>>", file);
        return -1;
    }

    return 0;
}

/* The following two routines and data structure are from the ever-brilliant
   Rob Leslie.
*/

struct audio_dither {
  mad_fixed_t error[3];
  mad_fixed_t random;
};

/*
* NAME:        prng()
* DESCRIPTION: 32-bit pseudo-random number generator
*/
static inline
unsigned long prng(unsigned long state)
{
  return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}

/*
* NAME:        audio_linear_dither()
* DESCRIPTION: generic linear sample quantize and dither routine
*/
inline
signed long audio_linear_dither(unsigned int bits, mad_fixed_t sample,
                                struct audio_dither *dither)
{
  unsigned int scalebits;
  mad_fixed_t output, mask, random;

  enum {
    MIN = -MAD_F_ONE,
    MAX =  MAD_F_ONE - 1
  };

  /* noise shape */
  sample += dither->error[0] - dither->error[1] + dither->error[2];

  dither->error[2] = dither->error[1];
  dither->error[1] = dither->error[0] / 2;

  /* bias */
  output = sample + (1L << (MAD_F_FRACBITS + 1 - bits - 1));

  scalebits = MAD_F_FRACBITS + 1 - bits;
  mask = (1L << scalebits) - 1;

  /* dither */
  random  = prng(dither->random);
  output += (random & mask) - (dither->random & mask);

  dither->random = random;

  /* clip */
  if (output > MAX) {
    output = MAX;

    if (sample > MAX)
      sample = MAX;
  }
  else if (output < MIN) {
    output = MIN;

    if (sample < MIN)
      sample = MIN;
  }

  /* quantize */
  output &= ~mask;

  /* error feedback */
  dither->error[0] = sample - output;

  /* scale */
  return output >> scalebits;
}

enum mad_flow output(void *data,
                     struct mad_header const *header,
                     struct mad_pcm *pcm)
{
    register int nsamples = pcm->length;
    register int n;
    mad_fixed_t const *left_ch = pcm->samples[0], *right_ch = pcm->samples[1];
    
    static unsigned char stream[MP3_FRAME_SIZE    /* size of mp3 frame */
                               * 2              /* 2 channels */
                               * 2              /* 2 bytes / sample */
#ifdef USE_MAD_RESAMPLE    
                               * 4              /* allow for a 4x increase due
                                                   to sample rate conversion */
#endif /* USE_MAD_RESAMPLE */
        ];
#ifdef USE_MAD_RESAMPLE    
    static mad_fixed_t (*resample_buf)[2 /*channels*/]
        [MP3_FRAME_SIZE   /* size of mp3 frame */
        * 4] = 0;       /* alow for 4x increase due to sample rate conversion
                         */
    static struct resample_state resampler_left, resampler_right;
#endif /* USE_MAD_RESAMPLE */
    static unsigned int prev_rate = 0;    /* sample rate of previous call */
    static unsigned int output_rate = 0;  /* rate at which we set device */
    static int channels = 0;
    static struct audio_dither dither;
    register int err = 0;

    register char * ptr = stream;
    register signed int sample;
    register mad_fixed_t tempsample;

    /* We need to know information about the file before we can open the playdevice
       in some cases. So, we do it here. 
       If either of these cases are true, most likely it is because the song
       has changed
     */
    if (!playdevice)
    {
#ifdef USE_MAD_RESAMPLE        
        if ( resample_buf ) {
            resample_finish(&resampler_left);
            resample_finish(&resampler_right);
            free(resample_buf);
            resample_buf = 0;
        }    
#endif
        /* TIVO */
        channels = MAD_NCHAN_WOPT(options, header);
        prev_rate = output_rate = header->samplerate;
        err = open_ao_playdevice(header);
    }
    /* TIVO */
    else if ((channels != MAD_NCHAN_WOPT(options, header)
              || prev_rate != header->samplerate) 
             && playdevice_is_live() ) 
    {
        if (playdevice) {
            ao_close(playdevice);
        }
        /* TIVO */
        channels = MAD_NCHAN_WOPT(options, header);
        prev_rate = output_rate = header->samplerate;

#ifdef USE_MAD_RESAMPLE
        if ( resample_buf) {
            resample_finish(&resampler_left);
            resample_finish(&resampler_right);
            free(resample_buf);
            resample_buf = 0;
        }    
#endif /* USE_MAD_RESAMPLE */    
        
        err = open_ao_playdevice(header);
    }
    
    if ( (err &&
          header->samplerate != 48000 &&
          header->samplerate != 44100 &&
          header->samplerate != 32000 )
#ifdef FORCE_RESAMPLE
         || (!resample_buf)
#endif
        )
    {
#ifdef USE_MAD_RESAMPLE        
        struct mad_header tmp_header;
        /* If we got here, let's assume that it was because the device couldn't
         * be opened because of a sample rate problem (XXXThis isn't
         * necessarily true!), so try one of 32Khz, 44.1Khz, or 48Khz and do
         * samplerate conversion below.  We'll use the following mappings:
         *  8Khz, 16Khz         : 32Khz
         *  12Khz, 24Khz        : 48Khz
         *  22.05Khz, 11.025Khz : 44.1Khz
         */
        /* TIVO */
        channels = MAD_NCHAN_WOPT(options, header);
        
        if (header->samplerate == 8000 ||   // 8Khz
            header->samplerate == 16000) {  // 16Khz
            output_rate = 32000;
        } else if (header->samplerate == 12000 || // 12Khz
                   header->samplerate == 24000) { // 24Khz
            output_rate = 48000;
        } else if (header->samplerate == 22050 || // 22.05Khz
                   header->samplerate == 11025) { // 11.025Khz
            output_rate = 44100;
        } else {   // assume 44.1Khz...quality will not be as good
            output_rate = 44100;
        }
        // In order to get around the "constness" of the header structure, we
        // copy it here, modify the rate, and open the device
        tmp_header = *header;
        tmp_header.samplerate = output_rate;
        if (playdevice) {
            ao_close(playdevice);
        }
        fprintf(stderr, "Trying opening device with rate %d, and using sample "
                "rate conversion\n", output_rate);
        err = open_ao_playdevice(&tmp_header);

        if ( err )
        {
            fprintf(stderr, 
                    "Couldn't open device even with sample rate conversion\n");
            return MAD_FLOW_BREAK;
        }
        else
        {
            if (resample_buf) {
                resample_finish(&resampler_left);
                resample_finish(&resampler_right);
            }
            else {
                resample_buf = malloc(sizeof(*resample_buf));
                if ( ! resample_buf) {
                    mpg321_error("Not enough memory to resample");
                    // Couldn't resample between these two rates
                    return MAD_FLOW_BREAK;
                }
            }
            
            if (-1 == resample_init(&resampler_left, header->samplerate, output_rate) ||
                -1 == resample_init(&resampler_right, header->samplerate, output_rate) ) 
            {
                mpg321_error("Couldn't resample between %d and %d",
                             header->samplerate, output_rate);
                if (resample_buf) {
                    free(resample_buf);
                    resample_buf = 0;
                }
                // Couldn't resample between these two rates
                return MAD_FLOW_BREAK;
            }
            
            fprintf(stderr, "mpg321: sample rate converting from %d to %d\n",
                    header->samplerate, output_rate);
        }
#endif /* USE_MAD_RESAMPLE */        
    }
    
    if (err) {
        return MAD_FLOW_BREAK;
    }

#ifdef USE_MAD_RESAMPLE
    if ( resample_buf )
    {
        nsamples = resample_block(&resampler_left,
                                  pcm->length,
                                  pcm->samples[0], 
                                  (*resample_buf)[0]);
        left_ch = (*resample_buf)[0];
        
        if ( pcm->channels == 2 )
        {
            resample_block(&resampler_right,
                           pcm->length, 
                           pcm->samples[1], 
                           (*resample_buf)[1]);
            right_ch = (*resample_buf)[1];
        }
        else right_ch = 0;
    }
#endif /* USE_MAD_RESAMPLE */    

    {
        long nTotalSamples = nsamples;
        n = nsamples;
        if (pcm->channels == 2)
        {
            while (n--)
            {
#ifdef SOFTWARE_VOLUME
                tempsample = (mad_fixed_t)((*left_ch++ *
                                            (double)options.volume)/MAD_F_ONE);
#else
                tempsample = *left_ch++;
#endif
                sample = (signed int) audio_linear_dither(16, tempsample, &dither);

#ifndef WORDS_BIGENDIAN
                *ptr++ = (unsigned char) (sample >> 0);
                *ptr++ = (unsigned char) (sample >> 8);
#else
                *ptr++ = (unsigned char) (sample >> 8);
                *ptr++ = (unsigned char) (sample >> 0);
#endif
            
#ifdef SOFTWARE_VOLUME
                tempsample = (mad_fixed_t)((*right_ch++ *
                                            (double)options.volume)/MAD_F_ONE);
#else            
                tempsample = *right_ch++;
#endif            
                sample = (signed int) audio_linear_dither(16, tempsample, &dither);
#ifndef WORDS_BIGENDIAN
                *ptr++ = (unsigned char) (sample >> 0);
                *ptr++ = (unsigned char) (sample >> 8);
#else
                *ptr++ = (unsigned char) (sample >> 8);
                *ptr++ = (unsigned char) (sample >> 0);
#endif
            }

            nTotalSamples = nsamples * 4;
        }
    
        else if (options.opt & MPG321_FORCE_STEREO)
        {
            while (n--)
            {
#ifdef SOFTWARE_VOLUME            
                tempsample = (mad_fixed_t)((*left_ch++ *
                                            (double)options.volume)/MAD_F_ONE); 
#else
                tempsample = *left_ch++;
#endif            
                sample = (signed int) audio_linear_dither(16, tempsample, &dither);
            
                /* Just duplicate the sample across both channels. */
#ifndef WORDS_BIGENDIAN
                *ptr++ = (unsigned char) (sample >> 0);
                *ptr++ = (unsigned char) (sample >> 8);
                *ptr++ = (unsigned char) (sample >> 0);
                *ptr++ = (unsigned char) (sample >> 8);
#else
                *ptr++ = (unsigned char) (sample >> 8);
                *ptr++ = (unsigned char) (sample >> 0);
                *ptr++ = (unsigned char) (sample >> 8);
                *ptr++ = (unsigned char) (sample >> 0);
#endif
            }

            nTotalSamples = nsamples * 4;
        }
        
        else /* Just straight mono output */
        {
            while (n--)
            {
#ifdef SOFTWARE_VOLUME            
                tempsample = (mad_fixed_t)((*left_ch++ *
                                            (double)options.volume)/MAD_F_ONE);
#else 
                tempsample = *left_ch++;
#endif
                sample = (signed int) audio_linear_dither(16, tempsample, &dither);
            
#ifndef WORDS_BIGENDIAN
                *ptr++ = (unsigned char) (sample >> 0);
                *ptr++ = (unsigned char) (sample >> 8);
#else
                *ptr++ = (unsigned char) (sample >> 8);
                *ptr++ = (unsigned char) (sample >> 0);
#endif
            }
            nTotalSamples = nsamples * 2;
        }
        
        ao_play(playdevice, stream, nTotalSamples);
    }

    return MAD_FLOW_CONTINUE;        
}

enum mad_flow decode_filter(void *data, 
                            struct mad_stream const *stream,
			    struct mad_frame *frame)
{
    buffer *playbuf = data;
    enum mad_flow mf = MAD_FLOW_CONTINUE;

#ifndef OLD_PAUSE_IMPL    
    if(options.opt & MPG321_REMOTE_PLAY)
    {
        /* We might have to stop if the user inputs something */
        if ((mf = remote_get_input_nowait(playbuf)))
            return mf;
    }
    if ( status == MPG321_PAUSED )
    {
        mpg321_printf("%s\n", "WAITING FOR UNPAUSE");
        
        mf = remote_get_input_wait(playbuf);
    }
#endif
    
    return mf;
}
